import atexit as _atexit

from .event_emitter import Event, EventEmitter
from .application_listener import ApplicationListener
from .component_listener import ComponentListener

__all__ = ["Event", "ApplicationListener", "ComponentListener"]

emitter = EventEmitter()

# register exit callback
_atexit.register(emitter.clear)


import yaml
import re
from autoboot.util.capacity import Capacity
from .env_process import load_env_file
from .yaml_parser import load_yaml_file, get_yml_value
from .yaml_constructors import (
  Loader, 
  construct_include, 
  environment_vars_constructor, 
  duration_constructor,
  capacity_constructor
)


yaml.add_constructor(u'!include', construct_include, Loader)
yaml.add_constructor(u'!env', environment_vars_constructor, Loader)
yaml.add_constructor(u'!ts', duration_constructor, Loader)
yaml.add_constructor(u'!cap', capacity_constructor, Loader)

pattern = re.compile(r'^(\d+h)?\s*(\d+m)?\s*(\d+s)?\s*(\d+ms)?\s*$')
yaml.add_implicit_resolver(u'!ts', pattern, Loader=Loader)
yaml.add_implicit_resolver(u'!cap', re.compile(pattern=Capacity.pattern), Loader=Loader)

__all__ = ["load_env_file", "load_yaml_file", "get_yml_value"]




import re

class Capacity(object):
  """Capacity represent resolver."""
  
  pattern = r'^(\d+)([KMGT])B$'
  
  def __init__(self, representation: str) -> None:
    parts = re.match(Capacity.pattern, representation)
    if parts is None:
      raise ValueError(f'Invalid capacity representation: {representation}')
    self.value = int(parts[1])
    self.unit = parts[2]
    
  def __str__(self) -> str:
    return f'{self.value}{self.unit}'
  
  def __repr__(self) -> str:
    return f'Capacity({self.value}, {self.unit})'
  
  def __eq__(self, other: object) -> bool:
    if not isinstance(other, Capacity):
      return False
    return self.value == other.value and self.unit == other.unit
  
  def to_bytes(self) -> int:
    """
    Converts capacity to bytes.
    """
    
    if self.unit == 'K':
      return self.value * 1024
    elif self.unit == 'M':
      return self.value * 1024 * 1024
    elif self.unit == 'G':
      return self.value * 1024 * 1024 * 1024
    elif self.unit == 'T':
      return self.value * 1024 * 1024 * 1024 * 1024 
    else:
      raise ValueError(f'Invalid capacity unit: {self.unit}')
    
  def to_kb(self) -> int:
    """
    Converts capacity to KiB.
    """
    
    if self.unit == 'K':
      return self.value
    elif self.unit == 'M':
      return self.value * 1024
    elif self.unit == 'G':
      return self.value * 1024 * 1024
    elif self.unit == 'T':
      return self.value * 1024 * 1024 * 1024
    else:
      raise ValueError(f'Invalid capacity unit: {self.unit}')
      
  def to_mb(self) -> float:
    """
    Converts capacity to MiB.
    """
    
    if self.unit == 'K':
      return self.value / 1024.0
    elif self.unit == 'M':
      return self.value
    elif self.unit == 'G':
      return self.value * 1024
    elif self.unit == 'T':
      return self.value * 1024 * 1024
    else:
      raise ValueError(f'Invalid capacity unit: {self.unit}')
  
  def to_gb(self) -> float:
    """
    Converts capacity to GiB.
    """
    
    if self.unit == 'K':
      return self.value / 1024.0 / 1024
    elif self.unit == 'M':
      return self.value / 1024.0
    elif self.unit == 'G':
      return self.value
    elif self.unit == 'T':
      return self.value * 1024
    else:
      raise ValueError(f'Invalid capacity unit: {self.unit}')
    
  def to_tb(self) -> float:
    """
    Converts capacity to TiB.
    """
    
    if self.unit == 'K':
      return self.value / 1024.0 / 1024 / 1024
    elif self.unit == 'M':
      return self.value / 1024.0 / 1024
    elif self.unit == 'G':
      return self.value / 1024.0
    elif self.unit == 'T':
      return self.value
    else:
      raise ValueError(f'Invalid capacity unit: {self.unit}')
  
  



from typing import Any


def get_value_by(dict: dict[str, Any], keypath: str) -> str | None:
    """
    Get the value of a keypath in a dictionary.

    Args:
        dict (dict): The dictionary to get the value from.
        keypath (str): The keypath to get the value from.

    Returns:
        str: The value of the keypath.
    """
    keys = keypath.split(".")
    value = dict
    for key in keys:
        value = value.get(key)
        if value is None:
            return None
    return value



import os
import wrapt
from functools import wraps
from typing import Callable, TypeVar
from autoboot.applications import AutoBoot
from autoboot.annotation.component import component
from autoboot.process import get_yml_value

R = TypeVar("R")

def env(keypath: str) -> int | float | str | dict | list | None:
  """get env value from keypath, support find in `.env` and `.yml`."""
  
  if "." in keypath:
    value = get_yml_value(AutoBoot.get_config_data(), keypath=keypath)
    if not value:
      return None
    return value
  else:
    value = os.getenv(keypath)
    if value:
      return value
    value = AutoBoot.get_config_data().get(keypath)
    return value if value else None
    

def value(keypath: str) -> R:
  """get value from config file."""
  
  @wrapt.decorator
  def decorator(fn: Callable[..., R], instance, args, kwargs) -> R:
    return env(keypath) or fn(*args, **kwargs)
  return decorator
      

def value_component(keypath: str) -> R:
  """get value from config file and cache as component."""
  
  @wrapt.decorator
  @component(f"env[{keypath}]")
  def decorator(fn: Callable[..., R], instance, args, kwargs) -> R:
    return env(keypath) or fn(*args, **kwargs)
  return decorator

def static_property(keypath: str) -> R:
  """Decorate as a static method, get value from config file and cache as component.

  Args:
      keypath (str): config key name

  Returns:
      R: config value
  """ 
  
  # @staticmethod not work with @wrapt.decorator
  def wrapper(fn: Callable[..., R]) -> R:
    @staticmethod
    @component(f"env[{keypath}]")
    @value(keypath)
    @wraps(fn)
    def decorator(*args, **kwargs):
        return fn(*args, **kwargs)
    return decorator
  return wrapper



from autoboot.annotation.env import static_property


class ApplicationProperties:
  
  @static_property("autoboot.application.name")
  def app_name() -> str:
    return "app"

  @static_property("autoboot.application.module")
  def module() -> str:
    return "api"
  
  @static_property("autoboot.application.scan_listener_packages")
  def scan_listener_packages() -> list[str]:
    pass
  
  @static_property("autoboot.application.log.dir")
  def log_dir() -> str:
    return "logs"
  
  @static_property("autoboot.application.log.max_size")
  def log_max_size() -> str:
    return "100 MB"

  @static_property("autoboot.application.log.retention")
  def log_retention() -> str:
    return "30 days"


import json
from typing import Type, TypeVar
from pydantic import BaseModel

T = TypeVar('T', bound=BaseModel)

def stringify(model: T) -> str:
  return model.json(exclude_none=True, by_alias=True)

def parse(model_class: Type[T], data: str) -> T:
  dict = json.loads(data)
  return model_class(**dict)



from functools import wraps
from time import perf_counter
from typing import Callable, TypeVar
from loguru import logger
from autoboot.applications import AutoBoot
from autoboot.logging import Logging

R = TypeVar("R")

def log_time(fn: Callable[..., R]) -> R:
  """Debug function take time after run finish."""
  
  @wraps(fn)
  def decorator(*args, **kwargs) -> R:
    start = perf_counter()
    try:
      result = fn(*args, **kwargs)
      end = perf_counter()
      duration = end - start
      AutoBoot.logger.debug(f"{fn.__name__} took {duration:4f} seconds to execute")
      return result
    except Exception as e:
      AutoBoot.logger.exception(f"exception happen with msg: {e}")
  return decorator

def log_catch(user_case: str):
  """Add catch error log with loguru.

  A decorator add on function type, it used logger config.

  Args:
  user_case: set current user case
  """
  
  def wrapper(fn: Callable[..., R]) -> Callable[..., R]:
    @logger.catch
    @wraps(fn)
    def decorator(*args, **kwargs) -> R:
      Logging.catch_file_rotation_logger(user_case)
      return fn(*args, **kwargs)
    return decorator
  return wrapper



from dataclasses import dataclass
from autoboot.event import emitter, Event
from loguru import logger

@dataclass(slots=True)
class PayOrder:
  no: str
  
@dataclass(slots=True)
class RepayOrder:
  no: str
  

def test_event_emitter():
  # send event name action
  emitter.emit(action="pay_action", event=Event("pay order: 1001"))
  # send event with with generic type
  emitter.emit(event=Event(PayOrder("1001")))
  
@emitter.on("pay_action")
def received_payment(event: Event[str]):
  logger.info("received_payment")
  assert(event.data == "pay order: 1001")
  
@emitter.on_event
def received_pay(event: Event[PayOrder]):
  logger.info("received_pay")
  assert(event.data == PayOrder("1001"))

@emitter.on_event
def received_repay(event: Event[RepayOrder]):
  logger.info("received_repay")
  assert(event.data == RepayOrder("1001"))



import os
from typing import Any

import yaml
from .yaml_constructors import Loader
from autoboot.util.object_navigator import get_value_by


def load_yaml_file(dir: str, file_name: str) -> dict[str, Any]:
  file_path: str
  if os.path.exists(os.path.join(dir, file_name + ".yml")):
    file_path = f"{file_name}.yml"
  else:
    file_path = f"{file_name}.yaml"
  with open(os.path.join(dir, file_path), 'r') as f:
    return yaml.load(f, Loader)
  
def get_yml_value(config_data: dict[str, Any], keypath: str) -> Any:
  value = get_value_by(config_data, keypath)
  if not value:
    return None
  if isinstance(value, str) and value.startswith("env(") and value.endswith(")"):
    return os.getenv(value[4:-1])
  return value




from concurrent.futures import ThreadPoolExecutor
from autoboot.meta import Component
from autoboot import ServerProperties
from typing import Callable, Any


class ThreadPoolTask(Component):
  """The ThreadPoolTask class is extends Component which means is a singleton instance."""
  
  component_name = "thread_pool_task"
  
  def __init__(self):
    self.executor = ThreadPoolExecutor(max_workers=ServerProperties.thread_max_workers())
    
  def worker_size(self):
    return len(self.executor._threads)
  
  def queue_size(self):
    return self.executor._work_queue.qsize()
  
  def is_overflow(self):
    return (self.queue_size() + 1) > ServerProperties.thread_max_queue()
    
  def submit(self, fn: Callable[..., Any]):
    rejected = ServerProperties.thread_rejected()    
    if self.is_overflow():
      if rejected == "abort":
        raise RuntimeError("ThreadPoolTask: Too many task executing")
      elif rejected == "discard":
        return
    return self.executor.submit(fn)
  
  def shutdown(self):
    self.executor.shutdown(wait=True)
    
  def __call__(self, fn: Callable[..., Any]):
    return self.submit(fn)
    
  def __enter__(self):
    return self
  
  def __exit__(self, exc_type, exc_val, exc_tb):
    self.shutdown()
    
  def __del__(self):
    self.shutdown()
    
  def __repr__(self):
    return f"ThreadPoolTask(size={self.worker_size()}, queue={self.queue_size()})"

  def __str__(self):
    return f"ThreadPoolTask(size={self.worker_size()}, queue={self.queue_size()})"



import os
import json
import yaml
from typing import Any, IO
from durations import Duration
from autoboot.util.capacity import Capacity


class Loader(yaml.SafeLoader):
  """YAML Loader for custom constructor."""

  def __init__(self, stream: IO) -> None:
    """Initialise Loader."""

    try:
      self._root = os.path.split(stream.name)[0]
    except AttributeError:
      self._root = os.path.curdir

    super().__init__(stream)


def construct_include(loader: Loader, node: yaml.Node) -> Any:
  """Include file referenced at node."""

  filename = os.path.abspath(os.path.join(loader._root, loader.construct_scalar(node)))
  extension = os.path.splitext(filename)[1].lstrip(".")

  with open(filename, "r") as f:
    if extension in ("yaml", "yml"):
      return yaml.load(f, Loader)
    elif extension in ("json",):
      return json.load(f)
    else:
      return "".join(f.readlines())


def environment_vars_constructor(loader: Loader, node: yaml.Node) -> Any:
  value = loader.construct_scalar(node)
  if '$' in value:
    # support search env var as $VAR or ${VAR}
    return os.path.expandvars(value)
  return os.getenv(value)


def duration_constructor(loader: Loader, node: yaml.Node) -> Any:
  value = loader.construct_scalar(node)
  return int(Duration(value).to_miliseconds())

def capacity_constructor(loader: Loader, node: yaml.Node) -> Any:
  value = loader.construct_scalar(node)
  return int(Capacity(value).to_bytes())



from enum import Enum
from pydantic import BaseModel
from autoboot.util import json_model

def test_parse():
  class Address(BaseModel):
    city: str
    
  class UserType(Enum):
    NORMAL = 0
    VIP = 1
  
  class User(BaseModel):
    name: str
    age: int
    address: Address
    type: UserType
  
  data = '{"name": "John", "age": 30, "type": 0, "address": {"city": "New York"}}'
  user = json_model.parse(User, data)
  assert(user==User(name="John", age=30, type=UserType.NORMAL, address=Address(city="New York")))


import threading
import weakref
from typing import Union, TypeVar, Generic

T = TypeVar("T")

class ThreadLocal(Generic[T]): 
  
  def __init__(self) -> None:
    self.localMap = weakref.WeakValueDictionary()
    self.lock = threading.RLock()
    
  
  def set(self, key: Union[str, int], value: T) -> None:
    with self.lock:
      self.localMap[key] = value
  
  def get(self, key: str | int) -> T:
    with self.lock:
      return self.localMap.get(key)
    
  def clear(self, key: str | int) -> None:
    with self.lock:
      self.localMap.pop(key, None)


import inspect
from typing import Any, Callable, get_args


def has_property(obj: Any, prop: str, is_own: bool) -> bool:
  """
  Checks if a property is present in an object.

  Args:
    obj(Any): The object to check.
    prop(str): The property to check.
    is_own(bool): Whether to check the object's own properties or all properties.
    
  Returns:
    bool: True if the property is present, False otherwise.
  """
  if is_own:
    # properties from object.__dict__
    return prop in vars(obj)
  else:
    # properties from self or super class
    return hasattr(obj, prop)
  
def get_method_params(fn: Callable[..., Any]):
  return inspect.signature(fn).parameters

def get_param_generic(param: inspect.Parameter, index: int):
  return get_args(param.annotation)[index]

def get_generic_type(fn: Callable[..., Any], param_name: str, index: int):
  params = get_method_params(fn)
  event_param = params.get(param_name)
  return get_param_generic(event_param, index) if event_param else None

def match_param_type(param: inspect.Parameter, type: Any):
  return param.annotation == type



from functools import wraps
from time import sleep
from typing import TypeVar, Callable
from result import Ok, Err, Result
from loguru import logger
  
R = TypeVar("R")
  
def retry(tries: int, delay: float, exceptions: type[Exception] | tuple[type[Exception], ...] = BaseException):
  """
  Retry a function if it fails.
  """
  
  def wrapper(fn: Callable[..., R]) -> Callable[..., Result[R, str]]:
    @wraps(fn)
    def decorator(*args, **kwargs) -> Result[R, str]:
      for i in range(tries):
        try:
          return Ok(fn(*args, **kwargs))
        except exceptions as e:
          logger.exception(f"Retrying {fn.__name__} after exception:")
          if i < tries - 1:
            sleep((i * delay) ** i)
          else:
            return Err(str(e))
    return decorator
  return wrapper




class ComponentMetaclass(type):
  """
  Metaclass for all components.
  """
  
  def __init__(cls, name, bases, attrs):
    super(ComponentMetaclass, cls).__init__(name, bases, attrs)
    
    # get component name from class attrs['component_name']
    if name != 'Component':
      cls.__component_name__ = attrs['component_name'] \
        if 'component_name' in attrs and attrs['component_name'] else name

  def __new__(cls, name, bases, attrs):
    return super(ComponentMetaclass, cls).__new__(cls, name, bases, attrs)



from typing import Self
from autoboot.annotation.component import comp_from_cache
from .component_metaclass import ComponentMetaclass


class Component(metaclass=ComponentMetaclass):
  """Cache component and register it in IoC."""

  def __new__(cls) -> Self:
    """Create a new instance of the component."""
    _parent = super()
    return comp_from_cache(cls.__component_name__, lambda: _parent.__new__(cls))
  



import os
from dotenv import dotenv_values

def load_env_file(dir: str, env_name: str):
  """load env file with input opt arg: -e <name>"""
  
  _load_env_file(f"{dir}/.env")
  if env_name:
    _load_env_file(f"{dir}/.env.{env_name}")

def _load_env_file(filepath: str):
  env_kvs = dotenv_values(filepath)
  for k, v in env_kvs.items():
    os.environ[k] = v
      





from .component import component
from .env import env, value, value_component, static_property
from .log import log_time, log_catch


__all__ = ["component", "env", "value", "value_component", "static_property", "log_time", "log_catch"]



import abc
from typing import TypeVar, Generic
from .app_plugin_metaclass import AppPluginMetaclass

T = TypeVar('T')

class AppPlugin(Generic[T], metaclass=AppPluginMetaclass):
  """An abstract class used to extends for add to framework."""
  
  @classmethod
  def get_context(cls) -> T:
    """Don't override this class method which used by the internal mechanism of the framework."""
    from autoboot import AutoBoot
  
    return AutoBoot.get_runner(cls.__ctx__)
  
  @classmethod
  def get_context_name(cls) -> str:
    """Don't override this class method which used by the internal mechanism of the framework."""
    return cls.__ctx__
  
  @abc.abstractmethod
  def install(self) -> T:
    """Install plugin and return runner context."""
    pass
  
  @abc.abstractmethod
  def env_prepared(self) -> None:
    """Invoke after env vars prepared. 
    
    it should be noted that current runner context not created.
    """
    pass
  
  @abc.abstractmethod
  def app_started(self) -> None:
    """Invoke after application started success."""
    pass


from __future__ import annotations  # Postponed Evaluation of Annotations (PEP 563) was unusable, it requires a __future__ import.

import os
import sys
import loguru
from typing import TypeVar
from autoboot.args import Env
from loguru import logger

LoggingType = TypeVar("LoggingType", bound="Logging")

class Logging:
  """
  Logging class wrapper with loguru.
  """
  
  _instance: LoggingType = None
  
  def __init__(self, log_dir: str, app_name: str, module: str, max_size: str, retention: str, env_name: str):
    self.log_dir = log_dir
    self.app_name = app_name
    self.module = module
    self.max_size = max_size
    self.retention = retention
    self.logger = self.console_logger() if env_name == Env.DEV.value else self.file_rotation_logger()
    Logging._instance = self
    
  def console_logger(self) -> loguru.Logger:
    logger.configure(
      handlers=[{
        "sink": sys.stdout,
        "format": "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>",
        "level": "DEBUG",
        "colorize": True,
        "backtrace": True,
      }])
    return logger
    
  def file_rotation_logger(self) -> loguru.Logger:
    log_app_dir = os.path.join(self.log_dir, self.app_name)
    if not os.path.exists(log_app_dir):
      os.makedirs(log_app_dir)
    logger.configure(
      handlers=[{
        "sink": f"{log_app_dir}/{self.module}-{{time:YYYY-MM-DD}}.log",
        "rotation": self.max_size,
        "retention": self.retention,
        "level": "INFO",
        "format": "{time:YYYY-MM-DD HH:mm:ss.SSS} | {level: <8} | {name}:{module}:{function}:{line} - {message}",
        "colorize": False,
        "enqueue": True,
        "backtrace": True,
      }])
    return logger
  
  @classmethod
  def catch_file_rotation_logger(cls, user_case: str):
    log_app_dir = os.path.join(cls._instance.log_dir, cls._instance.app_name)
    if not os.path.exists(log_app_dir):
      os.makedirs(log_app_dir)
    logger.configure(extra={"user_case": user_case})
    logger.add(
      f"{log_app_dir}/{cls._instance.module}-error-{{time:YYYY-MM-DD}}.log",
      rotation=cls._instance.max_size,
      retention=cls._instance.retention,
      level="ERROR",
      format="{time:YYYY-MM-DD HH:mm:ss.SSS} | {level: <8} | {name}:{module}:{function}:{line} -> [{extra[user_case]}] - {message}",
      catch=True,
      enqueue=True,
      backtrace=True
    )
      





from dataclasses import dataclass
from typing import TypeVar, Generic, Callable
from autoboot.util.type_reflect import get_generic_type

T = TypeVar('T')

@dataclass(slots=True)
class Event(Generic[T]):
  data: T
  source: any = None
  

class EventEmitter(object):
  """
  EventEmitter is a class that allows you to register event listeners and
  emit events.
  """

  def __init__(self):
    self._listeners = {}

  def on(self, action="default"):
    """
    Register a callback for an event with action name.
    """
    
    def decorator(fn: Callable[..., None], *args, **kwargs) -> None:
      self._listeners.setdefault(action, []).append(fn)
    return decorator
  
  def on_event(self, fn: Callable[..., None]):
    """
    Register a callback for an event with event generic type.
    """
    
    self._listeners.setdefault("default", []).append(fn)
    

  def emit(self, event: Event[T], action = "default"):
    """
    Emit event with action
    """
    
    check_type = action == "default"
    for f in self._listeners.get(action, []):
      if not check_type:
        f(event)
      else:
        generic_type = get_generic_type(f, "event", 0)
        if generic_type and isinstance(event.data, generic_type):
          f(event)
      
  def clear(self):
    self._listeners.clear()


from .application_properties import ApplicationProperties
from .server_properties import ServerProperties
from .applications import AutoBoot, AutoBootConfig

VERSION  = "0.8.0"

__all__ = ["AutoBoot", "AutoBootConfig", "ApplicationProperties", "ServerProperties"]





from typing import Type

from autoboot import AutoBoot
from autoboot.event import ApplicationListener, ComponentListener


def Listener(cls):
  
  if issubclass(cls, ApplicationListener):
      AutoBoot.instance().addListener(Type.__new__(cls))
  elif issubclass(cls, ComponentListener):
    AutoBoot.instance().addComponentListener(Type.__new__(cls))



from typing import Any, Protocol, runtime_checkable


@runtime_checkable
class ApplicationListener(Protocol):
  """Global application listener protocol."""
  
  def on_env_prepared(self, config: dict[str, Any]):...
  
  def on_started(self):...


from result import Ok, Err
from loguru import logger
from autoboot.annotation.effect import retry

def test_retry():
  
  @retry(tries=3, delay=1)
  def request():
    # test throw exception
    # 1 / 0
    pass
    

  match request():
    case Ok(value):
      logger.info(f"val = {value}")
    case Err(value):
      logger.error(f"err = {value}")



import loguru
from dataclasses import dataclass
from typing import Any, Callable, Optional, Tuple, TypeVar, Self
from autoboot.args import get_env_name
from autoboot.logging import Logging
from autoboot.plugin import AppPlugin
from autoboot.process import load_env_file, load_yaml_file
from autoboot.event import ApplicationListener
from autoboot.event import ComponentListener

# define AutoBoot class type
AppType = TypeVar("AppType", bound="AutoBoot")

R = TypeVar("R")

@dataclass(slots=True)
class AutoBootConfig:
  """AutoBoot global config"""
  
  # show detail log
  verbose: bool = False
  # config dir path which relative to the current project
  config_dir: str = "."
  # yaml config name, it support `.yml` and `.yaml`
  yml_config: str = "autoboot"
  
  

class AutoBoot(object):
  """The root application context which use of single containers is designed for simplicity and microservice."""
  
  _instance = None
  _init_flag = False
  
  def __new__(cls, config: Optional[AutoBootConfig]) -> Self:
    # create instance only once
    if cls._instance is None:
        cls._instance = super().__new__(cls)
    return cls._instance
  
  def __init__(self, config: Optional[AutoBootConfig] = AutoBootConfig()) -> None:
    if AutoBoot._init_flag is True:
      return
    self.config = config
    self._app_plugins: list[AppPlugin] = []
    self._components: list[Tuple[str, Any]] = []
    self._component_listeners: list[ComponentListener] = []
    self._runners: dict[str, Any] = {}
    self._listeners: list[ApplicationListener] = []
    self._config_data = dict[str, Any]
    AutoBoot.logger: loguru.Logger = None
    AutoBoot._init_flag = True
    
  @classmethod
  def instance(cls):
    """Return current application context instance."""
    return cls._instance
  
  @classmethod
  def get_config_data(cls) -> dict[str, Any]:
    """Get yaml config dict data."""
    return cls.instance()._config_data
  
  @staticmethod
  def get_runner(name: str):
    """Get plugin runner with name."""
    return AutoBoot.instance()._runners.get(name)
     
  @property
  def app_plugins(self) -> list[AppPlugin]:
    """Get application plugins."""
    return self._app_plugins
  
  @app_plugins.setter
  def app_plugins(self, app_plugins: list[AppPlugin]) -> None:
    self._app_plugins = app_plugins
  
  @property
  def components(self) -> list[Tuple[str, Any]]:
    """Ioc container for find depends."""
    return self._components
    
  def apply(self, app_plugin: AppPlugin) -> Self:
    """Apply plugin in application context."""
    self._app_plugins.append(app_plugin)
    return self
    
  def addListener(self, listener: ApplicationListener) -> Self:
    self._listeners.append(listener)
    return self
  
  def addComponentListener(self, listener: ComponentListener) -> Self:
    self._component_listeners.append(listener)
    return self
  
  def run(self, expose: Callable[..., R] = None):
    """Application context run entry."""
    
    # config environment variables
    env_name = get_env_name()
    
    # load .env
    load_env_file(self.config.config_dir, env_name)
    
    # load yml
    self._config_data = load_yaml_file(self.config.config_dir, self.config.yml_config)
    
    # config logger
    from autoboot.application_properties import ApplicationProperties
    AutoBoot.logger = Logging(
      app_name=ApplicationProperties.app_name(),
      module=ApplicationProperties.module(),
      log_dir=ApplicationProperties.log_dir(),
      max_size=ApplicationProperties.log_max_size(),
      retention=ApplicationProperties.log_retention(),
      env_name=env_name
    ).logger
    
    AutoBoot.logger.info(f"application finish load env: {env_name}")
    
    for ap in self.app_plugins:
      ap.env_prepared()
      
    for listener in self._listeners:
      listener.on_env_prepared(self._config_data)
      
    for ap in self.app_plugins:
      runner = ap.install()
      self._runners[ap.get_context_name()] = runner
      
    if ApplicationProperties.scan_listener_packages():
      for package in ApplicationProperties.scan_listener_packages():
        __import__(package)

    for ap in self.app_plugins:
      ap.app_started()
      
    for listener in self._listeners:
      listener.on_started()
      
    AutoBoot.logger.info("application started!")
    
    if expose:
      return expose()




class AppPluginMetaclass(type):
  
  def __init__(cls, name, bases, attrs):
    super(AppPluginMetaclass, cls).__init__(name, bases, attrs)

    if name != 'AppPlugin':
      cls.__ctx__ = attrs['context_name'] \
        if 'context_name' in attrs and attrs['context_name'] else name


from functools import wraps
from typing import Callable, TypeVar
from loguru import logger
from autoboot.applications import AutoBoot

R = TypeVar("R")

def comp_from_cache(name: str, comp_factory: Callable[..., R]):
  """
  Get component from cache
  """
 
  app = AutoBoot.instance()
  # return cached component if match
  for key, comp in app.components:
    if key == name:
      return comp
  comp = comp_factory()
  
  for listener in app._component_listeners:
    listener.after_instantiation(name, comp)
    
  # caching component with name
  if comp:
    app.components.append((name, comp))
    (AutoBoot.logger or logger).info(f"component cached: {name}")
  return comp

def component(name: str):
  """register as a component which cached with name for next time get"""
  def wrapper(fn: Callable[..., R]) -> Callable[..., R]:
    @wraps(fn)
    def decorator(*args, **kwargs) -> R:
      return comp_from_cache(name, lambda: fn(*args, **kwargs))
    return decorator
  return wrapper


""" define component using class style
class Component:
  def __init__(self, name: str):
    self.name = name
    
  def __call__(self, fn: Callable[..., R]) -> Callable[..., R]:
    @wraps(fn)
    def decorator(*args, **kwargs) -> R:
      app: AutoBoot = AutoBoot.instance()
      # return cached component if match
      for key, comp in app.components:
        if key == self.name:
          return comp
      result = fn(*args, **kwargs)
      # caching component with name
      if result:
        app.components.append((self.name, result))
        (AutoBoot.logger or logger).info(f"component cached: {self.name}")
      return result
    return decorator 
"""




from autoboot.annotation.env import static_property

class ServerProperties:

  @static_property("server.threads.max")
  def thread_max_workers() -> int:
    """The max workers is 31"""
    return 28
  
  @static_property("server.threads.queue_size")
  def thread_max_queue() -> int:
    """The max queue is 1000"""
    return 1000
  
  @static_property("server.threads.rejected")
  def thread_rejected() -> str:
    """Available reject strategy: abort, discard, wait"""
    return "wait"


import os
import sys
import getopt
import argparse
from enum import Enum


class Env(Enum):
  DEV = "dev"
  TEST = "test"
  PROD = "prod"
  

def get_env_name() -> str:
  parser = argparse.ArgumentParser(description="Startup")
  parser.add_argument("-e", required=False, dest="env_name", help="env name")
  args = parser.parse_known_args()
  if args[0].env_name:
    return args[0].env_name
  else:
    env_name = os.getenv("ENV_NAME")
    return env_name if env_name else Env.DEV.value
    
  # ops = sys.argv[1:]
  # if "-e" in ops:
  #   opts, args = getopt.getopt(ops, "e:") # "e:"中的`:`则表示-e参数后面应该带一个值
  #   for op, value in opts:
  #     if op == "-e":
  #       return value
  # return Env.DEV.value